"""
Run context

We can either be in param-init stage,
or in the main training loop,
or forwarding loop.
"""

from __future__ import annotations


__all__ = ["RunCtx", "get_current_run_ctx"]


def get_current_run_ctx() -> RunCtx:
    """
    :return: current run context, see :class:`RunCtx`
    """
    pass  # TODO...


class RunCtx:
    """
    We can either be in param-init stage,
    or in the main training loop,
    or forwarding loop.

    In training, we expect that some loss is being defined via mark_as_loss().
    In forwarding, we expect that some output is being defined via mark_as_output().
    """

    # TODO ...



#!/usr/bin/env python3

"""
This is being used by some tests.
"""

from __future__ import annotations
import sys
import os
import tempfile
import argparse


_my_dir = os.path.dirname(os.path.abspath(__file__))
_base_dir = os.path.dirname(_my_dir)
assert os.path.exists("%s/rnn.py" % _base_dir)


def setup(old_style=False, target_package_name="returnn"):
    """
    Setup env/sys.path such that `import returnn` works.

    :param bool old_style: make it an old-style setup
    :param str target_package_name:
    """
    print("Setup for importing RETURNN as framework/package.")
    tmp_env_path_dir = tempfile.mkdtemp()
    print("Temp dir:", tmp_env_path_dir)
    if old_style:
        print("Old-style setup!")
        src_dir = _base_dir
    else:
        src_dir = "%s/returnn" % _base_dir
    os.symlink(src_dir, "%s/%s" % (tmp_env_path_dir, target_package_name))
    sys.path.insert(0, tmp_env_path_dir)

    print("Import %s module/package." % target_package_name)
    if target_package_name == "returnn":
        # noinspection PyUnresolvedReferences
        import returnn
    else:
        __import__(target_package_name)

    print("Setup better_exchook.")
    if target_package_name == "returnn":
        if old_style:
            # noinspection PyUnresolvedReferences
            from returnn import better_exchook

            better_exchook.install()
        else:
            from returnn.util import better_exchook

            better_exchook.install()
    else:
        # Should always work. But only use for the fallback.
        __import__(target_package_name).better_exchook.install()


def test_TaskSystem_Pickler():
    from returnn.util.task_system import Pickler
    from returnn.util.task_system import BytesIO

    stream = BytesIO()
    pickler = Pickler(stream)
    obj = {"foo": "bar"}  # some dummy dict
    pickler.dump(obj)


def test_old_style_import_crnn_TFUtil():
    """
    This assumes that you use ``--returnn-package-name "crnn"`.
    """
    # noinspection PyUnresolvedReferences
    import crnn.TFUtil

    # noinspection PyUnresolvedReferences
    print("TF:", crnn.TFUtil.tf_version_tuple())


def test_old_style_import_TFUtil():
    # noinspection PyUnresolvedReferences
    import returnn.TFUtil

    # noinspection PyUnresolvedReferences
    print("TF:", returnn.TFUtil.tf_version_tuple())


if __name__ == "__main__":
    print("RETURNN as framework/package.")
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("py_eval_str", nargs="*")
    arg_parser.add_argument("--old-style", action="store_true")
    arg_parser.add_argument("--returnn-package-name", default="returnn")
    args = arg_parser.parse_args()
    setup(old_style=args.old_style, target_package_name=args.returnn_package_name)
    for py_eval_str in args.py_eval_str:
        print("-" * 20)
        print("Run:", py_eval_str)
        print("-" * 20)
        eval(py_eval_str)
        print("-" * 20)
        print("Ok.")
        print("-" * 20)



"""
Random number generator utilities
"""

from __future__ import annotations
from typing import Optional, Union, Sequence
import numpy
from returnn.tensor import Tensor, Dim
from returnn.util.basic import NotSpecified


# noinspection PyUnusedLocal
def random(
    *,
    shape: Sequence[Dim],
    distribution: str,
    mean: Optional[Union[int, float, Tensor]] = NotSpecified,
    stddev: Optional[Union[int, float, Tensor]] = NotSpecified,
    bound: Optional[Union[int, float, Tensor]] = NotSpecified,
    minval: Optional[Union[int, float, Tensor]] = NotSpecified,
    maxval: Optional[Union[int, float, Tensor]] = NotSpecified,
    dtype: str = NotSpecified,
    seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = NotSpecified,
    algorithm: Optional[str] = NotSpecified,
    explicit_state: Optional[Tensor] = NotSpecified,
    auto_update_state: Optional[bool] = NotSpecified,
    static: Optional[bool] = NotSpecified,
) -> Tensor:
    """
    Generates random numbers from uniform or normal or truncated normal distribution.

    This uses the TensorFlow stateless random ops internally, i.e. all the state handling is explicit.
    The state var can be explicitly provided and initialized via :class:`RandomStateInitLayer`,
    or when not provided it will be automatically created.

    There are two possible distinct use cases:

    - For any randomness in the model, e.g. dropout. So each ``session.run`` step will produce a new random number
      and advance the random state.
    - To initialize parameters via the config, using :class:`VariableLayer` with the ``init_by_layer`` option.
      This will only be called once when initializing the parameters.
      For this use case, we do not want to keep a random state var.
      You can just pass ``static=False``.
      Alternatively you could also pass the output of a :class:`RandomStateInitLayer` as ``state``.

    :param Sequence[Dim] shape:
    :param str distribution: "uniform", "normal" or "truncated_normal"
    :param int|float|Tensor|None mean:
    :param int|float|Tensor|None stddev:
    :param int|float|Tensor|None bound: for uniform, defining the range [-bound, bound)
    :param int|float|Tensor|None minval: for uniform
    :param int|float|Tensor|None maxval: for uniform
    :param str dtype:
    :param int|list[int]|numpy.ndarray|None seed: If not given, uses self.network.random.randint,
      i.e. then it is controlled by the global seed setting, and every layer would get its own seed.
      If you specify it explicitly, make sure every :class:`RandomLayer` uses a different seed,
      otherwise you would get the same random numbers everywhere.
    :param str|tf.random.Algorithm|None algorithm: see :class:`RandomStateInitLayer`
    :param Tensor|None explicit_state: You can pass the state explicitly here.
      If not given, will be created automatically, and updated automatically.
      You could pass a :class:`VariableLayer` with initial value via :class:`RandomStateInitLayer`,
      or directly a :class:`RandomStateInitLayer`.
      If auto_update_state is True, it must be a variable,
      and every time a new random number is created, this variable is updated.
      Otherwise (default), it will not be updated automatically.
    :param bool|None auto_update_state: only used when you pass an explicit state
    :param bool|None static: if no state at all should be used. it just relies on the seed then.
    :return: layer
    """
    raise NotImplementedError  # TODO



"""
RETURNN frontend (returnn.frontend) tests
"""


import returnn.frontend as rf
from returnn.tensor import Tensor, Dim


def _demo_run(model: rf.Module):
    model  # noqa  # TODO
    raise NotImplementedError  # TODO


def test_simple_net_linear():
    class _Net(rf.Module):
        def __init__(self, in_dim: Dim, out_dim: Dim):
            super().__init__()
            self.linear = rf.Linear(in_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            """
            Forward
            """
            return self.linear(x)

    _demo_run(_Net(Dim(7, name="in"), Dim(13, name="out")))



"""
DType helpers
"""

from __future__ import annotations


__all__ = ["get_default_int_dtype", "get_default_float_dtype"]


_default_int_dtype: str = "int32"
_default_float_dtype: str = "float32"


def get_default_int_dtype() -> str:
    """
    :return: default dtype for int
    """
    return _default_int_dtype


def get_default_float_dtype() -> str:
    """
    :return: default dtype for float
    """
    return _default_float_dtype
